# -*- coding: utf-8 -*-
'''

 MODULE:       hru-delin_init

 AUTHOR(S):    adapted from GRASS-HRU (ILMS) - JENA University
               by IRSTEA - Christine Barachet 

 PURPOSE:      Prepare files for the next steps of HRU Delineation
                   shapefile : selected gauges
                   rasters :   bounded DEM and others (geology, soils, landuse)
                               DEM, slope and aspect with values of reclassification
                               calculates oriented flows and subbasins

'''

import string, subprocess, os, shutil, sys, glob, types
import math
import ConfigParser
import grass.script as grass

from osgeo import gdal
from osgeo import ogr


'''
- cutting the rasters according to user selection
- done with GDAL which is not operating under GRASS environment
'''
def cutting_raster(raster,raster_out):
    global xmin_slect,ymax_slect,xmax_slect,ymin_slect
    cmd = ['gdal_translate', '-projwin', str(xmin_slect), str(ymax_slect), str(xmax_slect), str(ymin_slect), raster, raster_out]    
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.communicate()

#   -------------------------
def get_raster_bounds(layer):
#   -------------------------
    global xmin_slect,ymax_slect,xmax_slect,ymin_slect
    gdal_module = 'gdalinfo ' + layer
    p = subprocess.Popen(gdal_module +' | grep "Upper Left"', shell=True, stdout=subprocess.PIPE)
    upper_left = p.communicate()[0]
    xmin_slect = max(xmin_slect, float(upper_left.split(")")[0].split("(")[1].split(",")[0]))
    ymax_slect = min(ymax_slect, float(upper_left.split(")")[0].split("(")[1].split(",")[1]))
    p = subprocess.Popen(gdal_module +' | grep "Lower Right"', shell=True, stdout=subprocess.PIPE)
    lower_right = p.communicate()[0]
    xmax_slect = min(xmax_slect, float(lower_right.split(")")[0].split("(")[1].split(",")[0]))
    ymin_slect = max(ymin_slect, float(lower_right.split(")")[0].split("(")[1].split(",")[1]))
    
   
#   --------------------------
def get_polygon_bounds(shape):
#   --------------------------
    global xmin_slect,ymax_slect,xmax_slect,ymin_slect
    ogr_module = 'ogrinfo -al ' +shape
    p = subprocess.Popen(ogr_module +' | grep "Extent"', shell=True, stdout=subprocess.PIPE)
    extent = p.communicate()[0]
    xmin_slect = max(xmin_slect, float(extent.split(")")[0].split("(")[1].split(",")[0]))
    xmax_slect = min(xmax_slect, float(extent.split(")")[1].split("(")[1].split(",")[0]))
    ymin_slect = max(ymin_slect, float(extent.split(")")[0].split("(")[1].split(",")[1]))
    ymax_slect = min(ymax_slect, float(extent.split(")")[1].split("(")[1].split(",")[1]))
       
#   -------------
def get_coords():
#   -------------
    global xmin_slect,ymax_slect,xmax_slect,ymin_slect
    xmin_slect = max(xmin_slect, float(parms.get("surface", "west")))
    ymax_slect = min(ymax_slect, float(parms.get("surface", "north")))
    xmax_slect = min(xmax_slect, float(parms.get("surface", "east")))
    ymin_slect = max(ymin_slect, float(parms.get("surface", "south")))


''' 

 MAIN
 

'''
try:
    from subprocess import DEVNULL # Python 3.
except ImportError:
    #DEVNULL = open('cbtrace_step1', 'wb')
    DEVNULL = open(os.devnull, 'wb') 

print('---------- HRU-delin Step 1 started ---------------------------------------------')

# Get parameters from configuration file
parms = ConfigParser.ConfigParser(allow_no_value=True)
parms_file = "hrudelin_config.cfg"
parms.read(parms_file)
directory = parms.get("dir_in", "dir")
directory_out = parms.get("dir_out", "files")

# Initializing the variables for cutting up of layers
xmin_slect = ymin_slect =0.0
ymax_slect = xmax_slect =9999999999.9

dem = directory + '/' + str(parms.get("files_in", "dem"))
dem_name = str(parms.get("files_in", "dem")).split(".")[0]

selection = (parms.get("surface","selection"))
get_raster_bounds(dem)

if selection == 'total':
    pass
elif selection == 'polygon':
     get_polygon_bounds(directory + '/' + parms.get("surface", "polygon"))
elif selection == 'coords':
     get_coords()
else:
    print('==============================================================================')
    print('No surface specified in hrudelin_config or invalid parameter, full DEM applied')
    print('==============================================================================')

print('---------- HRU-delin Step 1 : Preparing the layers')
# Cutting up of the DEM
dem_cut = directory_out + '/step1_dem_cut.tif'    
cutting_raster(dem,dem_cut)

irr_rast = str(parms.get("irrigation", "irrig_rast"))
if irr_rast != '':
    irr_rast = directory + '/' + irr_rast
    layer_cut = directory_out + '/step1_irrigation.tif'
    cutting_raster(irr_rast,layer_cut)

# Cutting up of the gauges shapefile
gauges_shp = (directory + '/' + str(parms.get("files_in", "gauges")))
gauges_selected = str(directory_out + '/gauges_selected.shp')
## this section (to ensure intersection with DEM) 
#  does not work with point shapefiles with a single gauge
ogr_module = 'ogrinfo -al ' + gauges_shp
p = subprocess.Popen(ogr_module +' | grep "Extent"', shell=True, stdout=subprocess.PIPE)
extent = p.communicate()[0]
#xmin_g = max(xmin_slect, float(extent.split(")")[0].split("(")[1].split(",")[0]))
#xmax_g = min(xmax_slect, float(extent.split(")")[1].split("(")[1].split(",")[0]))
#ymin_g = max(ymin_slect, float(extent.split(")")[0].split("(")[1].split(",")[1]))
#ymax_g = min(ymax_slect, float(extent.split(")")[1].split("(")[1].split(",")[1]))
#cmd = ['ogr2ogr', '-spat', str(xmin_g), str(ymin_slect), str(xmax_g), str(ymax_g), gauges_selected, gauges_shp]
##dbg
# print(xmin_slect)
# print(ymin_slect)
# print(xmax_slect)
# print(ymax_slect)
cmd = ['ogr2ogr', '-spat', str(xmin_slect), str(ymin_slect), str(xmax_slect), str(ymax_slect), gauges_selected, gauges_shp]
# print(cmd)
p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
res_communicate = p.communicate()
if (len(res_communicate[0])!=0):
    print(res_communicate)

# Cutting up and save other rasters (landuse, soils, geology, ...)
for data in parms.items('data'):
	get_raster_bounds(directory + '/' + data[1])
for data in parms.items('data'):
	layer = (directory + '/' + data[1])
	layer_cut = (directory_out + '/step1_' + data[1])
	cutting_raster(layer,layer_cut)

dem_wk='dem_wk'
grass.run_command('g.proj', flags='c', georef=dem_cut, stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.in.gdal', flags='o', input=dem_cut, output=dem_wk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('g.region', flags='sp', rast = dem_wk, stdout = DEVNULL, stderr = DEVNULL)

# Fill the DEM (depressionless) 
dem_filled = elev = dem_wk
if (parms.get("demfill", "demfill") == 'yes'):
    print('---------- HRU-delin Step 1 : Filling the DEM')
    dem_filled = 'dem_filled'
    dem_temp = 'dem_temp'
    dir_temp = 'dir_temp'
    unfilled_areas = 'unfilled_areas'
    
    areas = 1
    while (areas > 0):
        grass.run_command('r.fill.dir', input=elev, elevation=dem_filled, direction=dir_temp, areas=unfilled_areas, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
        areas = int(grass.parse_command('r.info', flags='r', map=unfilled_areas, stdout = DEVNULL, stderr = DEVNULL)['max'])
        elev = dem_filled
 
# Slope and aspect derivation
print('---------- HRU-delin Step 1 : Derivating layers from MNT')       
dem_slope='dem_slope'
dem_aspect='dem_aspect'
grass.run_command('r.slope.aspect', elevation=dem_filled, slope=dem_slope, aspect=dem_aspect, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
         
# Reclassifying rasters
# DEM
recl_dem_file = 'reclass_rules_dem'

if (parms.get("reclass_dem", "rules_auto_dem")) == 'yes':
    step_reclass = int(parms.get("reclass_dem", "step_dem"))
    alti_min = float(grass.parse_command('r.info', flags='r', map=dem_filled, stdout = DEVNULL, stderr = DEVNULL)['min'])
    alti_max = float(grass.parse_command('r.info', flags='r', map=dem_filled, stdout = DEVNULL, stderr = DEVNULL)['max'])
    
    dem_rules = open(recl_dem_file,'w')
    new_max = math.floor(alti_max/step_reclass)*step_reclass+step_reclass
    if alti_min < 0:
		x = 1
		dem_rules.write(str(alti_min) + ':0:1:' + str(x) + '\n')
		new_min = 0
    else:
		x = 0
		new_min = math.floor(alti_min/step_reclass)*step_reclass
    while new_min < new_max:
		x += 1
		dem_rules.write(str(new_min) + ':' + str(new_min+step_reclass) + ':' + str(x) + ':' + str(x) + '\n')
		new_min += step_reclass
    dem_rules.close()
else:
   if not os.path.isfile(recl_dem_file):
        sys.exit('------------> File reclass_rules_dem is missing')
dem_rcl = 'dem_rcl'
cmd = 'cat '+recl_dem_file+' | r.recode input='+dem_filled+' output='+dem_rcl
p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
p.communicate()
# Save the rules to directory_out
cmd = 'mv '+recl_dem_file+' '+directory_out
subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

# Slope 
recl_slope_file = 'reclass_rules_slope'
if (parms.get("reclass_slope", "rules_auto_slope")) == 'yes':
    slp_min = float(grass.parse_command('r.info', flags='r', map=dem_slope, stdout = DEVNULL, stderr = DEVNULL)['min'])
    slp_max = float(grass.parse_command('r.info', flags='r', map=dem_slope, stdout = DEVNULL, stderr = DEVNULL)['max'])
    recl_slope_file = 'reclass_default_rules_slope'
    slp_rules = open(recl_slope_file,'r')
    slp_class = slp_rules.read()
    slp_rules.close()
    max_slp = str(slp_max)
    slp_class_new = slp_class.replace('max_slp', max_slp)
    max_class = str(int(math.ceil(slp_max)))
    slp_class = slp_class_new.replace('max_class', max_class)
    recl_slope_file = 'reclass_rules_slope'
    slp_rules = open(recl_slope_file,'w')
    slp_rules.write(slp_class)
    slp_rules.close()
else:
   if not os.path.isfile(recl_slope_file):
        sys.exit('------------> file reclass_rules_slope is missing')
        
slp_rcl='slp_rcl'
cmd = 'cat '+recl_slope_file+' | r.recode input='+dem_slope+' output='+slp_rcl
ps = subprocess.Popen(cmd,shell=True,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
ps.communicate()
cmd = 'mv '+recl_slope_file+' '+directory_out
subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

# Aspect 
recl_aspect_file = 'reclass_rules_aspect'
if (parms.get("reclass_aspect", "rules_auto_aspect")) == 'yes':
    recl_aspect = 'reclass_default_rules_aspect'
    subprocess.call('cp '+recl_aspect+' '+recl_aspect_file, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
else:
   if not os.path.isfile(recl_aspect_file):
        sys.exit('------------> File reclass_rules_aspect is missing')    
asp_rcl='asp_rcl'
cmd = 'cat '+recl_aspect_file+' | r.recode input='+dem_aspect+' output='+asp_rcl
ps = subprocess.Popen(cmd,shell=True,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
ps.communicate()
cmd = 'mv '+recl_aspect_file+' '+directory_out
subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

# Oriented flows generation
print('---------- HRU-delin Step 1 : Generating Oriented Flows')
min_size = int(parms.get('basin_min_size', 'size'))
streams = 'streams_wk'
subbasins = 'subbasins_wk'
halfbasins='halfbasins'
accum = 'accum_wk'
drain = 'drain_wk'
grass.run_command('r.watershed', elevation=dem_filled, stream=streams, basin=subbasins, half=halfbasins,accumulation=accum, drainage=drain, threshold=min_size, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Save rasters
print('---------- HRU-delin Step 1 : Saving the rasters')
raster_out = directory_out + '/' + 'step1_dem_filled.tif'
grass.run_command('r.out.gdal', input=dem_filled, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)   

raster_out = directory_out + '/' + 'step1_dem_reclass.tif'
grass.run_command('r.out.gdal', input=dem_rcl, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_slope.tif'  
grass.run_command('r.out.gdal', input=dem_slope, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_aspect.tif'  
grass.run_command('r.out.gdal', input=dem_aspect, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_slope_reclass.tif'  
grass.run_command('r.out.gdal', input=slp_rcl, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_aspect_reclass.tif'  
grass.run_command('r.out.gdal', input=asp_rcl, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_streams.tif' 
grass.run_command('r.out.gdal', input=streams, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_drain.tif' 
grass.run_command('r.out.gdal', input=drain, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_accum.tif' 
grass.run_command('r.out.gdal', input=accum, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_subbasins.tif' 
grass.run_command('r.out.gdal', input=subbasins, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

raster_out = directory_out + '/' + 'step1_halfbasins.tif' 
grass.run_command('r.out.gdal', input=halfbasins, output=raster_out, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
print('---------- HRU-delin Step 1 ended   ---------------------------------------------')
